% ----------------------
% newton.m
% ----------------------
% computes the root of a function using simple multidimensional Newton-Raphson Method
% in:           @f, x0
% in optional:  crit, showoutput, maxiters
% out:          x1, retcode, feval, iters
% ----------------------


function [x1, retcode, feval, iters] = newton(f, x0, varargin)

optargin = size(varargin,2);
varargin = cell2mat(varargin);

% Default values
crit        = 1e-9;
showoutput  = 0;
maxiters    = 100;

if optargin > 3
    error('Error in newton.m: Define at most 3 optional arguments.');
end

if optargin == 3
    crit        = varargin(1);
    showoutput  = varargin(2);
    maxiters    = varargin(3);
end

if optargin == 2
    crit        = varargin(1);
    showoutput  = varargin(2);
end

if optargin == 1
    crit        = varargin(1);
end

iters       = 0;
retcode     = 0;

delta       = 1;
feval       = f(x0);
nx0         = size(x0);
nfx0        = size(feval);
x1          = x0;

if nx0(2) > nx0(1)
    error('Error in newton.m: You have to specify the starting guess as a column vector.'); end

if nfx0(2) > nfx0(1)
    error('Error in newton.m: You have to specify the output of f() as a column vector.'); end


while (delta > crit && iters <= maxiters)

    fprime  = jacobian(f,x0);
    detjac  = det(fprime);
    if (abs(detjac) < 1e-12 || abs(detjac) >= inf)
        retcode = -1; break;
    end
    
    x1      = x0 - fprime\feval;
    
    feval   = f(x1);
    delta   = norm(feval,2);
    x0      = x1;
    iters   = iters+1;

end

% WARNINGS!
if (iters > maxiters)
    retcode = 1;
end
if (not(delta > -inf && delta < inf))
    retcode = 2;
end

if (showoutput == 1)
    if (retcode == 1); fprintf('WARNING: Maximum number of iterations reached!\n\n'); end
    if (retcode == 2); fprintf('WARNING: Undefined norm, complex solution!\n\n'); end
    if (retcode == -1); fprintf('WARNING: Near singluar Jacobian. Iter = %i, Det = %e\n\n',iters,det(fprime)); end
end
        
end

function jac = jacobian(f, x0)

epsilon = 1e-07;

n       = size(x0,1);
jac     = zeros(n,n);

feval   = f(x0);
dx0     = zeros(n,1); 

    for i=1:n
        
        dx      = dx0;
        dx(i)   = epsilon;

        jac(:,i) = (f(x0+dx)-feval)./epsilon;
    end
end


